# The subtrees of tree-level feynman diagrams
# the subtrees are always binary since we assume that all feynman vertices have 3 leaves
# the order of left and right don't matter because we are only interested in topologies

import Base.iterate

"""
    AbstractTree

Base type for trees, such as [`EmptyTree`](@ref) or [`Subtree`](@ref).

Implementations provide [`leaves`](@ref), returning the number of leaves of the tree.
"""
abstract type AbstractTree end

"""
    EmptyTree <: AbstractTree

A singleton type representing an empty tree, which has 1 leaf, see [`leaves`](@ref).
"""
struct EmptyTree <: AbstractTree end

"""
    Subtree <: AbstractTree

Representation of a specific subtree topology, with two subtrees of its own.
"""
struct Subtree <: AbstractTree
    left::AbstractTree
    right::AbstractTree
end

"""
    Forest

Representation of a forest containing all topologically unique trees with a given number of leaves.
Use [`iterate`](@ref) to iterate through the [`Subtree`](@ref)s.
"""
struct Forest
    leaves::Int

    function Forest(n::Int)
        @assert n > 0 "Can not create forest of trees with $n leaves, must be at least 1"
        new(n)
    end
end


"""
    leaves(tree::AbstractTree)

Returns the number of leaves or leaves of the tree.
"""
function leaves end

leaves(tree::EmptyTree) = 1

leaves(tree::Subtree) = leaves(tree.left) + leaves(tree.right)

leaves(forest::Forest) = forest.leaves
